@@ -2,7 +2,7 @@
  *
  *  OBEX Server
  *
- *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -25,6 +25,7 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
 #include <unistd.h>
 #include <string.h>
 #include <errno.h>
@@ -32,7 +33,6 @@
 #include <sys/socket.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/l2cap.h>
 #include <bluetooth/rfcomm.h>
 
 #include <openobex/obex.h>
@@ -40,13 +40,17 @@
 #include "bluetooth.h"
 #include "obexd.h"
 #include "obex.h"
+#include "obex-priv.h"
 #include "dbus.h"
 #include "logging.h"
 #include "btio.h"
 #include "service.h"
 
-#define TRANSFER_INTERFACE OPENOBEX_SERVICE ".Transfer"
-#define SESSION_INTERFACE OPENOBEX_SERVICE ".Session"
+#define OPENOBEX_MANAGER_PATH		"/"
+#define OPENOBEX_MANAGER_INTERFACE	OPENOBEX_SERVICE ".Manager"
+#define ERROR_INTERFACE			OPENOBEX_SERVICE ".Error"
+#define TRANSFER_INTERFACE		OPENOBEX_SERVICE ".Transfer"
+#define SESSION_INTERFACE		OPENOBEX_SERVICE ".Session"
 
 #define TIMEOUT 60*1000 /* Timeout for user response (miliseconds) */
 
@@ -409,7 +413,7 @@ failed:
 	return ret;
 }
 
-void register_record(struct server *server, gpointer user_data)
+void register_record(struct server *server)
 {
 	struct obex_service_driver *driver;
 	gchar *xml;
@@ -528,7 +532,8 @@ gboolean manager_init(void)
 
 	dbus_error_init(&err);
 
-	connection = g_dbus_setup_bus(DBUS_BUS_SESSION, OPENOBEX_SERVICE, &err);
+	connection = g_dbus_setup_private(DBUS_BUS_SESSION, OPENOBEX_SERVICE,
+									&err);
 	if (connection == NULL) {
 		if (dbus_error_is_set(&err) == TRUE) {
 			fprintf(stderr, "%s\n", err.message);
@@ -539,8 +544,11 @@ gboolean manager_init(void)
 	}
 
 	system_conn = g_dbus_setup_bus(DBUS_BUS_SYSTEM, NULL, NULL);
-	if (system_conn == NULL)
+	if (system_conn == NULL) {
+		dbus_connection_unref(connection);
+		connection = NULL;
 		return FALSE;
+	}
 
 	listener_id = g_dbus_add_service_watch(system_conn, "org.bluez",
 				name_acquired, name_released, NULL, NULL);
@@ -576,33 +584,9 @@ void manager_cleanup(void)
 	dbus_connection_unref(connection);
 }
 
-void emit_session_created(guint32 id)
+void manager_emit_transfer_started(struct obex_session *os)
 {
-	gchar *path = g_strdup_printf("/session%u", id);
-
-	g_dbus_emit_signal(connection, OPENOBEX_MANAGER_PATH,
-			OPENOBEX_MANAGER_INTERFACE, "SessionCreated",
-			DBUS_TYPE_OBJECT_PATH, &path,
-			DBUS_TYPE_INVALID);
-
-	g_free(path);
-}
-
-void emit_session_removed(guint32 id)
-{
-	gchar *path = g_strdup_printf("/session%u", id);
-
-	g_dbus_emit_signal(connection, OPENOBEX_MANAGER_PATH,
-			OPENOBEX_MANAGER_INTERFACE, "SessionRemoved",
-			DBUS_TYPE_OBJECT_PATH, &path,
-			DBUS_TYPE_INVALID);
-
-	g_free(path);
-}
-
-void emit_transfer_started(guint32 id)
-{
-	gchar *path = g_strdup_printf("/transfer%u", id);
+	gchar *path = g_strdup_printf("/transfer%u", os->cid);
 
 	g_dbus_emit_signal(connection, OPENOBEX_MANAGER_PATH,
 			OPENOBEX_MANAGER_INTERFACE, "TransferStarted",
@@ -612,7 +596,7 @@ void emit_transfer_started(guint32 id)
 	g_free(path);
 }
 
-void emit_transfer_completed(guint32 id, gboolean success)
+static void emit_transfer_completed(guint32 id, gboolean success)
 {
 	gchar *path = g_strdup_printf("/transfer%u", id);
 
@@ -625,7 +609,7 @@ void emit_transfer_completed(guint32 id, gboolean success)
 	g_free(path);
 }
 
-void emit_transfer_progress(guint32 id, guint32 total, guint32 transfered)
+static void emit_transfer_progress(guint32 id, guint32 total, guint32 transfered)
 {
 	gchar *path = g_strdup_printf("/transfer%u", id);
 
@@ -638,9 +622,9 @@ void emit_transfer_progress(guint32 id, guint32 total, guint32 transfered)
 	g_free(path);
 }
 
-void register_transfer(guint32 id, struct obex_session *os)
+void manager_register_transfer(struct obex_session *os)
 {
-	gchar *path = g_strdup_printf("/transfer%u", id);
+	gchar *path = g_strdup_printf("/transfer%u", os->cid);
 
 	if (!g_dbus_register_interface(connection, path,
 				TRANSFER_INTERFACE,
@@ -654,9 +638,13 @@ void register_transfer(guint32 id, struct obex_session *os)
 	g_free(path);
 }
 
-void unregister_transfer(guint32 id)
+void manager_unregister_transfer(struct obex_session *os)
 {
-	gchar *path = g_strdup_printf("/transfer%u", id);
+	gchar *path = g_strdup_printf("/transfer%u", os->cid);
+
+	/* Got an error during a transfer. */
+	if (os->object)
+		emit_transfer_completed(os->cid, os->offset == os->size);
 
 	g_dbus_unregister_interface(connection, path,
 				TRANSFER_INTERFACE);
@@ -728,9 +716,8 @@ static gboolean auth_error(GIOChannel *io, GIOCondition cond,
 	return FALSE;
 }
 
-int request_authorization(gint32 cid, int fd, const gchar *filename,
-			const gchar *type, gint32 length, gint32 time,
-			gchar **new_folder, gchar **new_name)
+int manager_request_authorization(struct obex_session *os, gint32 time,
+		gchar **new_folder, gchar **new_name)
 {
 	DBusMessage *msg;
 	DBusPendingCall *call;
@@ -739,8 +726,10 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	socklen_t addrlen;
 	gchar address[18];
 	const gchar *bda = address;
+	const gchar *filename = (os->name ? os->name : "");
+	const gchar *type = (os->type ? os->type : "");
 	gchar *path;
-	guint watch;
+	guint watch, fd;
 	gboolean got_reply;
 
 	if (!agent)
@@ -755,12 +744,13 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	memset(&addr, 0, sizeof(addr));
 	addrlen = sizeof(addr);
 
+	fd = OBEX_GetFD(os->obex);
 	if (getpeername(fd, (struct sockaddr *) &addr, &addrlen) < 0)
 		return -errno;
 
 	ba2str(&addr.rc_bdaddr, address);
 
-	path = g_strdup_printf("/transfer%d", cid);
+	path = g_strdup_printf("/transfer%d", os->cid);
 
 	msg = dbus_message_new_method_call(agent->bus_name, agent->path,
 					"org.openobex.Agent", "Authorize");
@@ -770,7 +760,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 			DBUS_TYPE_STRING, &bda,
 			DBUS_TYPE_STRING, &filename,
 			DBUS_TYPE_STRING, &type,
-			DBUS_TYPE_INT32, &length,
+			DBUS_TYPE_INT32, &os->size,
 			DBUS_TYPE_INT32, &time,
 			DBUS_TYPE_INVALID);
 
@@ -832,20 +822,6 @@ static void service_cancel(struct pending_request *pending)
 	g_dbus_send_message(system_conn, msg);
 }
 
-void obex_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
-{
-	struct server *server = user_data;
-
-	if (err) {
-		error("%s", err->message);
-		g_io_channel_shutdown(io, TRUE, NULL);
-		return;
-	}
-
-	if (obex_session_start(io, server) < 0)
-		g_io_channel_shutdown(io, TRUE, NULL);
-}
-
 static void pending_request_free(struct pending_request *pending)
 {
 	if (pending->call)
@@ -1002,9 +978,9 @@ gint request_service_authorization(struct server *server, GIOChannel *io,
 	return 0;
 }
 
-void register_session(guint32 id, struct obex_session *os)
+void manager_register_session(struct obex_session *os)
 {
-	gchar *path = g_strdup_printf("/session%u", id);
+	gchar *path = g_strdup_printf("/session%u", os->cid);
 
 	if (!g_dbus_register_interface(connection, path,
 				SESSION_INTERFACE,
@@ -1015,12 +991,22 @@ void register_session(guint32 id, struct obex_session *os)
 		return;
 	}
 
+	g_dbus_emit_signal(connection, OPENOBEX_MANAGER_PATH,
+			OPENOBEX_MANAGER_INTERFACE, "SessionCreated",
+			DBUS_TYPE_OBJECT_PATH, &path,
+			DBUS_TYPE_INVALID);
+
 	g_free(path);
 }
 
-void unregister_session(guint32 id)
+void manager_unregister_session(struct obex_session *os)
 {
-	gchar *path = g_strdup_printf("/session%u", id);
+	gchar *path = g_strdup_printf("/session%u", os->cid);
+
+	g_dbus_emit_signal(connection, OPENOBEX_MANAGER_PATH,
+			OPENOBEX_MANAGER_INTERFACE, "SessionRemoved",
+			DBUS_TYPE_OBJECT_PATH, &path,
+			DBUS_TYPE_INVALID);
 
 	g_dbus_unregister_interface(connection, path,
 				SESSION_INTERFACE);
@@ -1028,4 +1014,20 @@ void unregister_session(guint32 id)
 	g_free(path);
 }
 
+void manager_emit_transfer_progress(struct obex_session *os)
+{
+	emit_transfer_progress(os->cid, os->size, os->offset);
+}
 
+void manager_emit_transfer_completed(struct obex_session *os)
+{
+	emit_transfer_completed(os->cid, !os->aborted);
+}
+
+DBusConnection *obex_dbus_get_connection(void)
+{
+	if (connection == NULL)
+		return NULL;
+
+	return dbus_connection_ref(connection);
+}
