@@ -2,9 +2,8 @@
  *
  *  OBEX Server
  *
- *  Copyright (C) 2007-2008  Nokia Corporation
- *  Copyright (C) 2007-2008  Instituto Nokia de Tecnologia (INdT)
- *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2010  Nokia Corporation
+ *  Copyright (C) 2007-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -42,13 +41,14 @@
 #include <glib.h>
 
 #include <openobex/obex.h>
-#include <openobex/obex_const.h>
 
 #include "logging.h"
 #include "obex.h"
+#include "obex-priv.h"
 #include "dbus.h"
 #include "mimetype.h"
 #include "service.h"
+#include "btio.h"
 
 /* Default MTU's */
 #define DEFAULT_RX_MTU 32767
@@ -65,17 +65,53 @@ typedef struct {
 	guint16 mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
+static void os_set_response(obex_object_t *obj, int err)
+{
+	guint8 rsp;
+	guint8 lastrsp;
+
+	switch (err) {
+	case 0:
+		rsp = OBEX_RSP_CONTINUE;
+		lastrsp = OBEX_RSP_SUCCESS;
+		break;
+	case -EPERM:
+		rsp = OBEX_RSP_FORBIDDEN;
+		lastrsp = OBEX_RSP_FORBIDDEN;
+		break;
+	case -ENOENT:
+		rsp = OBEX_RSP_NOT_FOUND;
+		lastrsp = OBEX_RSP_NOT_FOUND;
+		break;
+	case -EBADR:
+		rsp = OBEX_RSP_BAD_REQUEST;
+		lastrsp = OBEX_RSP_BAD_REQUEST;
+		break;
+	case -EFAULT:
+		rsp = OBEX_RSP_SERVICE_UNAVAILABLE;
+		lastrsp = OBEX_RSP_SERVICE_UNAVAILABLE;
+		break;
+	case -EINVAL:
+		rsp = OBEX_RSP_NOT_IMPLEMENTED;
+		lastrsp = OBEX_RSP_NOT_IMPLEMENTED;
+		break;
+	default:
+		rsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
+		lastrsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
+	}
+
+	OBEX_ObjectSetRsp(obj, rsp, lastrsp);
+}
+
 static void os_reset_session(struct obex_session *os)
 {
 	if (os->object) {
+		os->driver->set_io_watch(os->object, NULL, NULL);
 		os->driver->close(os->object);
 		os->object = NULL;
-		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->current_folder) {
-			gchar *path;
-			path = g_build_filename(os->current_folder, os->name, NULL);
-			os->driver->remove(path);
-			g_free(path);
-		}
+		os->obj = NULL;
+		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path)
+			os->driver->remove(os->path);
 	}
 
 	if (os->name) {
@@ -90,6 +126,7 @@ static void os_reset_session(struct obex_session *os)
 		g_free(os->buf);
 		os->buf = NULL;
 	}
+
 	os->driver = NULL;
 	os->aborted = FALSE;
 	os->offset = 0;
@@ -113,9 +150,6 @@ static void obex_session_free(struct obex_session *os)
 
 	os_reset_session(os);
 
-	if (os->current_folder)
-		g_free(os->current_folder);
-
 	if (os->io)
 		g_io_channel_unref(os->io);
 
@@ -178,6 +212,9 @@ static void cmd_connect(struct obex_session *os,
 	guint hlen, newsize;
 	guint16 mtu;
 	guint8 hi;
+	const guint8 *target = NULL, *who = NULL;
+	guint target_size = 0, who_size = 0;
+	int err;
 
 	if (OBEX_ObjectGetNonHdrData(obj, &buffer) != sizeof(*nonhdr)) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -200,14 +237,21 @@ static void cmd_connect(struct obex_session *os,
 	os->cid = ++cid;
 
 	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
-		if (hi != OBEX_HDR_TARGET)
-			continue;
-
-		os->service = obex_service_driver_find(os->server->drivers,
-								hd.bs, hlen);
-		break;
+		switch (hi) {
+		case OBEX_HDR_WHO:
+			who = hd.bs;
+			who_size = hlen;
+			break;
+		case OBEX_HDR_TARGET:
+			target = hd.bs;
+			target_size = hlen;
+			break;
+		}
 	}
 
+	os->service = obex_service_driver_find(os->server->drivers,
+						target, target_size,
+						who, who_size);
 	if (os->service == NULL) {
 		error("Connect attempt to a non-supported target");
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -215,8 +259,26 @@ static void cmd_connect(struct obex_session *os,
 		return;
 	}
 
-	if (os->service->connect)
-		os->service->connect(obex, obj);
+	debug("Selected driver: %s", os->service->name);
+
+	if (!os->service->connect) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	}
+
+	os->service_data = os->service->connect(os, &err);
+	if (err == 0 && os->service->target) {
+		hd.bs = os->service->target;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_WHO, hd, 16,
+				OBEX_FL_FIT_ONE_PACKET);
+		hd.bq4 = os->cid;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_CONNECTION, hd, 4,
+				OBEX_FL_FIT_ONE_PACKET);
+	}
+
+	os_set_response(obj, err);
 }
 
 static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
@@ -254,6 +316,7 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 	obex_headerdata_t hd;
 	guint hlen;
 	guint8 hi;
+	int err;
 
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -318,7 +381,25 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 		}
 	}
 
-	os->service->get(obex, obj);
+	err = os->service->get(os, obj, os->service_data);
+	if (err == 0) {
+		if (os->size != OBJECT_SIZE_UNKNOWN) {
+			hd.bq4 = os->size;
+			OBEX_ObjectAddHeader(obex, obj,
+					OBEX_HDR_LENGTH, hd, 4, 0);
+		}
+
+		/* Add body header */
+		hd.bs = NULL;
+		if (os->size == 0)
+			OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
+					hd, 0, OBEX_FL_FIT_ONE_PACKET);
+		else
+			OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
+					hd, 0, OBEX_FL_STREAM_START);
+	}
+
+	os_set_response(obj, err);
 }
 
 static void cmd_setpath(struct obex_session *os,
@@ -326,6 +407,7 @@ static void cmd_setpath(struct obex_session *os,
 {
 	obex_headerdata_t hd;
 	guint32 hlen;
+	int err;
 	guint8 hi;
 
 	if (!os->service) {
@@ -367,25 +449,29 @@ static void cmd_setpath(struct obex_session *os,
 		break;
 	}
 
-	os->service->setpath(obex, obj);
+	err = os->service->setpath(os, obj, os->service_data);
+	os_set_response(obj, err);
 }
 
-int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
+int obex_stream_start(struct obex_session *os,
+		const gchar *filename, gpointer context)
 {
 	gint err;
 	gpointer object;
+	size_t size;
 
-	object = os->driver->open(filename, O_RDONLY, 0, size);
+	object = os->driver->open(filename, O_RDONLY, 0, context, &size, &err);
 	if (object == NULL) {
-		err = -errno;
-		error("open(%s): %s (%d)", filename, strerror(errno), errno);
+		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		goto fail;
 	}
 
+	os->driver->context = context;
 	os->object = object;
 	os->offset = 0;
+	os->size = size;
 
-	if (*size > 0)
+	if (size > 0)
 		os->buf = g_malloc0(os->tx_mtu);
 
 	return 0;
@@ -393,6 +479,7 @@ int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
 fail:
 	if (object)
 		os->driver->close(object);
+
 	return err;
 }
 
@@ -421,11 +508,13 @@ static gint obex_write_stream(struct obex_session *os,
 
 	len = os->driver->read(os->object, os->buf, os->tx_mtu);
 	if (len < 0) {
-		gint err = errno;
-		error("read(): %s (%d)", strerror(err), err);
+		error("read(): %s (%d)", strerror(-len), -len);
+		if (len == -EAGAIN)
+			return len;
+
 		g_free(os->buf);
 		os->buf = NULL;
-		return -err;
+		return len;
 	}
 
 	ptr = os->buf;
@@ -450,22 +539,21 @@ add_header:
 	return len;
 }
 
-gint os_prepare_put(struct obex_session *os)
+gint obex_prepare_put(struct obex_session *os, const gchar *filename)
 {
-	gchar *path;
 	gint len;
+	int err;
 
-	path = g_build_filename(os->current_folder, os->name, NULL);
-
-	os->object = os->driver->open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600,
-					(size_t *) &os->size);
+	os->object = os->driver->open(filename, O_WRONLY | O_CREAT | O_TRUNC,
+					0600, os->driver->context,
+					os->size != OBJECT_SIZE_UNKNOWN ?
+					(size_t *) &os->size : NULL, &err);
 	if (os->object == NULL) {
-		error("open(%s): %s (%d)", path, strerror(errno), errno);
-		g_free(path);
+		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		return -EPERM;
 	}
 
-	g_free(path);
+	os->path = filename;
 
 	if (!os->buf) {
 		debug("PUT request checked, no buffered data");
@@ -478,13 +566,11 @@ gint os_prepare_put(struct obex_session *os)
 
 		w = os->driver->write(os->object, os->buf + len, os->offset - len);
 		if (w < 0) {
-			gint err = errno;
-			error("write(%s): %s (%d)", path, strerror(errno),
-									errno);
-			if (err == EINTR)
+			error("write(%s): %s (%d)", filename, strerror(-w), -w);
+			if (w == -EINTR)
 				continue;
 			else
-				return -err;
+				return w;
 		}
 
 		len += w;
@@ -533,11 +619,10 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 
 		w = os->driver->write(os->object, buffer + len, size - len);
 		if (w < 0) {
-			gint err = errno;
-			if (err == EINTR)
+			if (w == -EINTR)
 				continue;
 			else
-				return -err;
+				return w;
 		}
 
 		len += w;
@@ -548,13 +633,52 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 	return 0;
 }
 
+static gboolean handle_async_io(gpointer object, int flags, int err,
+				gpointer user_data)
+{
+	struct obex_session *os = user_data;
+	int ret = 0;
+
+	if (err < 0) {
+		ret = err;
+		goto proceed;
+	}
+
+	if (flags & (G_IO_IN | G_IO_PRI))
+		ret = obex_write_stream(os, os->obex, os->obj);
+	else if (flags & G_IO_OUT)
+		ret = obex_read_stream(os, os->obex, os->obj);
+
+proceed:
+	switch (ret) {
+	case -EINVAL:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
+				OBEX_RSP_BAD_REQUEST);
+		break;
+	case -EPERM:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
+					OBEX_RSP_FORBIDDEN);
+		break;
+	default:
+		if (ret < 0)
+			OBEX_ObjectSetRsp(os->obj,
+					OBEX_RSP_INTERNAL_SERVER_ERROR,
+					OBEX_RSP_INTERNAL_SERVER_ERROR);
+		break;
+	}
+
+	OBEX_ResumeRequest(os->obex);
+
+	return FALSE;
+}
+
 static gboolean check_put(obex_t *obex, obex_object_t *obj)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
 	guint hlen;
 	guint8 hi;
-	int ret;
+	gint ret;
 
 	os = OBEX_GetUserData(obex);
 
@@ -634,25 +758,33 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		}
 	}
 
-	if (!os->service || !os->service->chkput)
+	if (!os->service->chkput)
 		goto done;
 
-	ret = os->service->chkput(obex, obj);
-	switch (ret) {
-	case 0:
-		break;
-	case -EINVAL:
+	if (!os->name) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
 				OBEX_RSP_BAD_REQUEST);
 		return FALSE;
+	}
+
+	ret = os->service->chkput(os, os->service_data);
+	switch (ret) {
+	case 0:
+		break;
 	case -EPERM:
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
+	case -EAGAIN:
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+		return TRUE;
 	default:
 		debug("Unhandled chkput error: %d", ret);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
 				OBEX_RSP_INTERNAL_SERVER_ERROR);
 		return FALSE;
+
 	}
 
 	if (os->size == OBJECT_SIZE_DELETE || os->size == OBJECT_SIZE_UNKNOWN) {
@@ -684,7 +816,7 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			return;
 	}
 
-	os->service->put(obex, obj);
+	os->service->put(os, os->service_data);
 }
 
 static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
@@ -699,30 +831,31 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
 		if (os->service->progress)
-			os->service->progress(obex, obj);
+			os->service->progress(os, os->service_data);
 		break;
 	case OBEX_EV_ABORT:
 		os->aborted = TRUE;
 		if (os->service->reset)
-			os->service->reset(obex);
+			os->service->reset(os, os->service_data);
 		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
 	case OBEX_EV_REQDONE:
 		switch (cmd) {
+		case OBEX_CMD_CONNECT:
+			break;
 		case OBEX_CMD_DISCONNECT:
 			OBEX_TransportDisconnect(obex);
 			break;
 		case OBEX_CMD_PUT:
 		case OBEX_CMD_GET:
 		case OBEX_CMD_SETPATH:
+		default:
 			os_session_mark_aborted(os);
 			if (os->service->reset)
-				os->service->reset(obex);
+				os->service->reset(os, os->service_data);
 			os_reset_session(os);
 			break;
-		default:
-			break;
 		}
 		break;
 	case OBEX_EV_REQHINT:
@@ -747,7 +880,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_REQCHECK:
 		switch (cmd) {
 		case OBEX_CMD_PUT:
-			if (os->service && os->service->put)
+			if (os->service)
 				check_put(obex, obj);
 			break;
 		default:
@@ -785,6 +918,11 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			OBEX_ObjectSetRsp(obj,
 				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 			break;
+		case -EAGAIN:
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io, os);
+			break;
 		default:
 			OBEX_ObjectSetRsp(obj,
 				OBEX_RSP_INTERNAL_SERVER_ERROR,
@@ -794,7 +932,21 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 		break;
 	case OBEX_EV_STREAMEMPTY:
-		obex_write_stream(os, obex, obj);
+		switch (obex_write_stream(os, obex, obj)) {
+		case -EPERM:
+			OBEX_ObjectSetRsp(obj,
+				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+			break;
+		case -EAGAIN:
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io,
+							os);
+			break;
+		default:
+			break;
+		}
+
 		break;
 	case OBEX_EV_LINKERR:
 		break;
@@ -825,7 +977,7 @@ static void obex_handle_destroy(gpointer user_data)
 	os = OBEX_GetUserData(obex);
 
 	if (os->service && os->service->disconnect)
-		os->service->disconnect(obex);
+		os->service->disconnect(os, os->service_data);
 
 	obex_session_free(os);
 
@@ -884,6 +1036,20 @@ failed:
 	return FALSE;
 }
 
+void obex_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
+{
+	struct server *server = user_data;
+
+	if (err) {
+		error("%s", err->message);
+		g_io_channel_shutdown(io, TRUE, NULL);
+		return;
+	}
+
+	if (obex_session_start(io, server) < 0)
+		g_io_channel_shutdown(io, TRUE, NULL);
+}
+
 gint obex_session_start(GIOChannel *io, struct server *server)
 {
 	struct obex_session *os;
@@ -892,9 +1058,8 @@ gint obex_session_start(GIOChannel *io, struct server *server)
 
 	os = g_new0(struct obex_session, 1);
 
-	os->service = obex_service_driver_find(server->drivers, NULL, 0);
-
-	os->current_folder = g_strdup(server->folder);
+	os->service = obex_service_driver_find(server->drivers, NULL, 0,
+						NULL, 0);
 	os->server = server;
 	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
 	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
@@ -944,16 +1109,86 @@ gint obex_tty_session_stop(void)
 	return 0;
 }
 
-struct obex_session *obex_get_session(gpointer object)
+const char *obex_get_name(struct obex_session *os)
 {
-	GSList *l;
+	return os->name;
+}
 
-	for (l = sessions; l; l = l->next) {
-		struct obex_session *os = l->data;
+ssize_t obex_get_size(struct obex_session *os)
+{
+	return os->size;
+}
 
-		if (os->object == object)
-			return os;
+const char *obex_get_type(struct obex_session *os)
+{
+	return os->type;
+}
+
+const char *obex_get_root_folder(struct obex_session *os)
+{
+	return os->server->folder;
+}
+
+guint16 obex_get_service(struct obex_session *os)
+{
+	return os->service->service;
+}
+
+gboolean obex_get_symlinks(struct obex_session *os)
+{
+	return os->server->symlinks;
+}
+
+const char *obex_get_capability_path(struct obex_session *os)
+{
+	return os->server->capability;
+}
+
+gboolean obex_get_auto_accept(struct obex_session *os)
+{
+	return os->server->auto_accept;
+}
+
+int obex_remove(struct obex_session *os, const char *path)
+{
+	if (os->driver == NULL)
+		return -EINVAL;
+
+	return os->driver->remove(path);
+}
+
+/* TODO: find a way to do this for tty or fix syncevolution */
+char *obex_get_id(struct obex_session *os)
+{
+	GError *gerr = NULL;
+	gchar address[18];
+	guint8 channel;
+
+	bt_io_get(os->io, BT_IO_RFCOMM, &gerr,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_CHANNEL, &channel,
+			BT_IO_OPT_INVALID);
+	if (gerr)
+		return NULL;
+
+	return g_strdup_printf("%s+%d", address, channel);
+}
+
+ssize_t obex_aparam_read(struct obex_session *os,
+		obex_object_t *obj, const guint8 **buffer)
+{
+	obex_headerdata_t hd;
+	guint8 hi;
+	guint32 hlen;
+
+	OBEX_ObjectReParseHeaders(os->obex, obj);
+
+	while (OBEX_ObjectGetNextHeader(os->obex, obj, &hi, &hd, &hlen)) {
+		if (hi == OBEX_HDR_APPARAM) {
+			*buffer = hd.bs;
+			return hlen;
+		}
 	}
 
-	return NULL;
+	return -EBADR;
 }
